/******************************************************************/
/* Linux IBM IPR microcode update utility                         */
/* Description: IBM Storage IOA Interface Specification (IPR)     */
/*              Linux microcode update utility                    */
/*                                                                */
/* (C) Copyright 2000, 2001                                       */
/* International Business Machines Corporation and others.        */
/* All Rights Reserved. This program and the accompanying         */
/* materials are made available under the terms of the            */
/* Common Public License v1.0 which accompanies this distribution.*/
/******************************************************************/

/*
 * $Header: /cvsroot/eserverio/iprutils/iprupdate.c,v 1.2.2.4 2004/10/14 18:31:22 bjking1 Exp $
 */

#include <unistd.h>
#include <sys/ioctl.h>
#include <scsi/sg.h>
#include <sys/stat.h>

#ifndef iprlib_h
#include "iprlib.h"
#endif

#include <sys/mman.h>

struct ipr_ioa_desc {
	u16 type;
	const char *desc;
};

struct ipr_ioa_desc ioa_desc [] = {
	{0x5702, "PCI-X Dual Channel Ultra320 SCSI Adapter [5702]"},
	{0x1974, "PCI-X Dual Channel Ultra320 SCSI Adapter [1974]"},
	{0x5703, "PCI-X Dual Channel Ultra320 SCSI RAID Adapter [5703]"},
	{0x1975, "PCI-X Dual Channel Ultra320 SCSI RAID Adapter [1975]"},
	{0x2780, "PCI-X Quad Channel Ultra320 SCSI RAID Adapter [2780]"},
	{0x5709, "SCSI RAID Enablement Card for PCI-X Dual Channel Ultra320 SCSI Integrated Controller [5709]"},
	{0x1976, "SCSI RAID Enablement Card for PCI-X Dual Channel Ultra320 SCSI Integrated Controller [1976]"},
	{0x570A, "PCI-X Dual Channel SCSI Integrated Controller (Adapter bus) [570A]"},
	{0x570B, "PCI-X Dual Channel SCSI Integrated Controller (Adapter bus) [570B]"}
};

static const char *get_ioa_desc(u16 type)
{
	int i;

	for (i = 0; i < sizeof(ioa_desc)/sizeof(ioa_desc[0]); i++) {
		if (type == ioa_desc[i].type)
			return ioa_desc[i].desc;
	}

	return NULL;
}

static void log_ucode_error(struct ipr_ioa *p_ioa)
{
	const char *desc = get_ioa_desc(p_ioa->ccin);

	if (desc) {
		syslog(LOG_ERR, "Could not find microcode file for IBM '%s'. "
		       "Please download the latest microcode from "
		       "http://techsupport.services.ibm.com/server/mdownload/download.html. "
		       "SCSI speeds will be limited to 80 MB/s until updated microcode is downloaded."IPR_EOL,
		       desc);
	} else {
		syslog(LOG_ERR, "Could not find microcode file for IBM %04X. "
		       "Please download the latest microcode from "
		       "http://techsupport.services.ibm.com/server/mdownload/download.html. "
		       "SCSI speeds will be limited to 80 MB/s until updated microcode is downloaded."IPR_EOL,
		       p_ioa->ccin);
	}
}

int open_dev(struct ipr_ioa *p_ioa, struct ipr_device *p_device);

int main(int argc, char *argv[])
{
    int rc = 0;
    struct stat ucode_stats;
    struct ipr_ioctl_cmd_internal ioa_cmd;
    struct ipr_inquiry_page0 page0_inq;
    struct ipr_inquiry_page3 page3_inq;
    char ucode_file[100];
    int do_download, ucode_fd, dasd_ucode_fd;
    struct image_header *p_image_hdr;
    struct ipr_dasd_ucode_header *p_dasd_image_hdr;
    u32 updated = 0;
    int fd, dev_fd, i, j, ioa_num;
    u32 force_download, force_devs, force_ioas;
    struct ipr_dasd_inquiry_page3 dasd_page3_inq;
    struct ipr_ioa *p_cur_ioa;
    char *p_dev_name;
    struct ipr_device *p_device;
    char config_file_name[64];
    u8 cdb[IPR_CCB_CDB_LEN];
    struct sense_data_t sense_data;
    int image_offset;
    void *p_dasd_image;
    int image_size;
    int device_update_blocked;
    int u320_disabled;
    u8 modification_level[4];

    openlog("iprupdate",
            LOG_PERROR | /* Print error to stderr as well */
            LOG_PID |    /* Include the PID with each error */
            LOG_CONS,    /* Write to system console if there is an error
                          sending to system logger */
            LOG_USER);

    force_download = force_devs = force_ioas = 0;

    if (argc > 1)
    {
        if (strcmp(argv[1], "--force") == 0)
            force_download = 1;
        else if (strcmp(argv[1], "--version") == 0)
        {
            printf("iprupdate: %s"IPR_EOL, IPR_VERSION_STR);
            exit(1);
        }
        else if (strcmp(argv[1], "--force-devs") == 0)
        {
            force_devs = 1;
        }
        else if (strcmp(argv[1], "--force-ioas") == 0)
        {
            force_ioas = 1;
        }
        else
        {
            printf("Usage: iprdate [options]"IPR_EOL);
            printf("  Options: --version    Print iprupdate version"IPR_EOL);
            exit(1);
        }
    }

    tool_init("iprupdate");

    check_current_config(false);

    /* Loop for all attached adapters */
    for (p_cur_ioa = p_head_ipr_ioa, ioa_num = 0;
         ioa_num < num_ioas;
         p_cur_ioa = p_cur_ioa->p_next, ioa_num++)
    {
        fd = open(p_cur_ioa->ioa.dev_name, O_RDWR | O_NONBLOCK);

        if (fd < 0)
        {
            syslog(LOG_ERR, "Cannot open %s. %m"IPR_EOL, p_cur_ioa->ioa.dev_name);
            continue;
        }

        /* Do a page 0 inquiry to the adapter to get the supported page codes */
        memset(&ioa_cmd, 0, sizeof(ioa_cmd));
        ioa_cmd.buffer = &page0_inq;
        ioa_cmd.buffer_len = sizeof(struct ipr_inquiry_page0);
        ioa_cmd.cdb[1] = 0x01;
        ioa_cmd.cdb[2] = 0x00;
        ioa_cmd.read_not_write = 1;

        rc = ipr_ioctl(fd, INQUIRY, &ioa_cmd);

        if (rc != 0)
        {
            syslog(LOG_ERR, "Inquiry to controller %s failed. %m"IPR_EOL, p_cur_ioa->ioa.dev_name);
            close(fd);
            continue;
        }

        /* Issue a page 3 inquiry to get the versioning information */
        memset(&ioa_cmd, 0, sizeof(ioa_cmd));
        ioa_cmd.buffer = &page3_inq;
        ioa_cmd.buffer_len = sizeof(struct ipr_inquiry_page3);
        ioa_cmd.cdb[1] = 0x01;
        ioa_cmd.cdb[2] = 0x03;
        ioa_cmd.read_not_write = 1;

        rc = ipr_ioctl(fd, INQUIRY, &ioa_cmd);

        if (rc != 0)
        {
            syslog(LOG_ERR, "Inquiry to controller %s failed. %m"IPR_EOL,
                   p_cur_ioa->ioa.dev_name);
            close(fd);
            continue;
        }

        do_download = 0;

        /* Parse the binary file */
        rc = find_ioa_firmware_image(p_cur_ioa, ucode_file);

        if (rc)
        {
		log_ucode_error(p_cur_ioa);
            close(fd);
            continue;
        }

        ucode_fd = open(ucode_file, O_RDONLY);

        if (ucode_fd == -1)
        {
            syslog(LOG_ERR, "Could not open firmware file %s. %m"IPR_EOL, ucode_file);
            close(fd);
            continue;
        }

        /* Get the size of the microcode image */
        rc = fstat(ucode_fd, &ucode_stats);

        if (rc != 0)
        {
            syslog(LOG_ERR, "Error accessing IOA firmware file: %s. %m"IPR_EOL, ucode_file);
            close(fd);
            close(ucode_fd);
            continue;
        }

        /* Map the image in memory */
        p_image_hdr = mmap(NULL, ucode_stats.st_size,
                           PROT_READ, MAP_SHARED, ucode_fd, 0);

        if (p_image_hdr == MAP_FAILED)
        {
            syslog(LOG_ERR, "Error mapping IOA firmware file: %s. %m"IPR_EOL, ucode_file);
            close(fd);
            close(ucode_fd);
            continue;
        }

        if (IPR_GET_FW_VERSION(*p_image_hdr) > IPR_GET_FW_VERSION(page3_inq))
            do_download = 1;

        u320_disabled = 0;

        /* check for mode page 28 changes before adapter reset */
        sprintf(config_file_name,"%x_%x", p_cur_ioa->ccin, p_cur_ioa->host_addr);
        if (RC_SUCCESS == ipr_config_file_valid(config_file_name))
        {
            ipr_set_page_28(p_cur_ioa, IPR_LIMITED_CONFIG,
                            do_download | force_download | force_ioas);
        }
        else
        {
            ipr_set_page_28_init(p_cur_ioa, IPR_LIMITED_CONFIG);
        }

        if ((do_download == 1) ||
            (force_download == 1) ||
            (force_ioas == 1))
        {
            syslog(LOG_NOTICE, "Updating IOA firmware on %s with %s"IPR_EOL,
                   p_cur_ioa->ioa.dev_name, ucode_file);

            /* Do the update if the two are different */
            /* Might want to fork off a process here so we can start
             on the next adapter. Would have to remember the child's PID
             so we could wait on it when we were done, though. This would also
             greatly complicate the load source case */

            memset(&ioa_cmd, 0, sizeof(ioa_cmd));
            ioa_cmd.buffer = (void *)((unsigned long)p_image_hdr + iprtoh32(p_image_hdr->header_length));
            ioa_cmd.buffer_len = ucode_stats.st_size - iprtoh32(p_image_hdr->header_length);
            ioa_cmd.read_not_write = 0;

            /* The write buffer here takes care of both the IOA shutdown and the
             reset of the adapter */

            rc = ipr_ioctl(fd, WRITE_BUFFER, &ioa_cmd);

            if (rc != 0)
                syslog(LOG_ERR, "Write buffer to %s failed. %m"IPR_EOL, p_cur_ioa->ioa.dev_name);
            else
            {
                /* Write buffer was successful. Now we need to see if all our devices
                 are available. The IOA may have been in braindead prior to the ucode
                 download, in which case the upper layer scsi device drivers do not
                 know anything about our devices. */

                check_current_config(false);

                for (j = 0; j < p_cur_ioa->num_devices; j++)
                {
                    /* If not a DASD, ignore */
                    if (!IPR_IS_DASD_DEVICE(p_cur_ioa->dev[j].p_resource_entry->std_inq_data))
                        continue;

                    /* See if we can open device */
                    dev_fd = open(p_cur_ioa->dev[j].dev_name, O_RDWR);

                    if ((dev_fd <= 1) && ((errno == ENOENT) || (errno == ENXIO)))
                    {
                        scan_device(p_cur_ioa->dev[j].p_resource_entry->resource_address,
                                    p_cur_ioa->host_num);
                    }
                    else
                        close(dev_fd);
                }
                updated = 1;
            }
        }

        /* Check DASD microcode */
        for (i = 0; i < p_cur_ioa->num_devices; i++)
        {
            p_device = &p_cur_ioa->dev[i];

            /* If not a DASD, ignore */
            if (!IPR_IS_DASD_DEVICE(p_cur_ioa->dev[i].p_resource_entry->std_inq_data) ||
                ((p_device->p_resource_entry->subtype != IPR_SUBTYPE_AF_DASD) &&
                 (p_device->p_resource_entry->subtype != IPR_SUBTYPE_GENERIC_SCSI)))
                continue;

            if (p_device->p_resource_entry->subtype == IPR_SUBTYPE_GENERIC_SCSI)
                p_dev_name = p_device->gen_name;
            else
                p_dev_name = p_cur_ioa->ioa.dev_name;

            /* Do a page 3 inquiry */
            dev_fd = open_dev(p_cur_ioa, p_device);

            if (dev_fd > 1)
            {
                if (p_device->p_resource_entry->subtype == IPR_SUBTYPE_GENERIC_SCSI) 
                {
                    memset(&dasd_page3_inq, 0, sizeof(dasd_page3_inq));
                    memset(cdb, 0, IPR_CCB_CDB_LEN);

                    /* Issue page 3 inquiry */
                    cdb[0] = INQUIRY;
                    cdb[1] = 0x01; /* EVPD */
                    cdb[2] = 0x03; /* Page 3 */
                    cdb[4] = sizeof(dasd_page3_inq);

                    rc = sg_ioctl(dev_fd, cdb, &dasd_page3_inq,
                                  sizeof(dasd_page3_inq),
                                  SG_DXFER_FROM_DEV, &sense_data, 30);

                    if (rc == CHECK_CONDITION)
                    {
                        syslog(LOG_ERR, "Page 3 inquiry to %02X%02X%02X%02X failed: %x %x %x"IPR_EOL,
                               p_device->p_resource_entry->host_no,
                               p_device->p_resource_entry->resource_address.bus,
                               p_device->p_resource_entry->resource_address.target,
                               p_device->p_resource_entry->resource_address.lun,
                               sense_data.sense_key & 0x0F,
                               sense_data.add_sense_code,
                               sense_data.add_sense_code_qual);
                        close(dev_fd);
                        continue;
                    }
                }
                else
                {
                    memset(&ioa_cmd, 0, sizeof(ioa_cmd));

                    ioa_cmd.buffer = &dasd_page3_inq;
                    ioa_cmd.buffer_len = sizeof(dasd_page3_inq);
                    ioa_cmd.cdb[1] = 0x01;
                    ioa_cmd.cdb[2] = 0x03;
                    ioa_cmd.read_not_write = 1;
                    ioa_cmd.device_cmd = 1;
                    ioa_cmd.resource_address = p_device->p_resource_entry->resource_address;

                    rc = ipr_ioctl(dev_fd, INQUIRY, &ioa_cmd);
                }

                if (rc != 0)
                {
                    syslog(LOG_ERR, "Inquiry to %02X%02X%02X%02X failed. %m"IPR_EOL,
                           p_device->p_resource_entry->host_no,
                           p_device->p_resource_entry->resource_address.bus,
                           p_device->p_resource_entry->resource_address.target,
                           p_device->p_resource_entry->resource_address.lun);
                    close(dev_fd);
                    continue;
                }

                device_update_blocked = 0;

                if (memcmp(p_device->p_resource_entry->std_inq_data.vpids.vendor_id, "IBM", 3)) {
                    close(dev_fd);
                    continue;
                }

                /* check if vendor id is not IBMAS400 */
                if (memcmp(p_device->p_resource_entry->std_inq_data.vpids.vendor_id, "IBMAS400", 8))
                {
                    if (p_device->opens)
                        device_update_blocked = 1;

                }

                do_download = 0;

                /* Parse the binary file */

                rc = find_dasd_firmware_image(p_device, dasd_page3_inq.load_id,
					      ucode_file, modification_level);

                if (rc == IPR_DASD_UCODE_NOT_FOUND)
                {
                    syslog_dbg(LOG_ERR, "Could not find firmware file %s. %m"IPR_EOL, prefix);
                    close(dev_fd);
                    rc = 0;
                    continue;
                }

                if (rc == IPR_DASD_UCODE_NO_HDR)
                    image_offset = 0;
                else
                    image_offset = sizeof(struct ipr_dasd_ucode_header);

                dasd_ucode_fd = open(ucode_file, O_RDONLY);

                if (dasd_ucode_fd < 0)
                {
                    syslog_dbg(LOG_ERR, "Could not open firmware file %s. %m"IPR_EOL, ucode_file);
                    close(dev_fd);
                    continue;
                }

                /* Get the size of the microcode image */
                rc = fstat(dasd_ucode_fd, &ucode_stats);

                if (rc != 0)
                {
                    syslog(LOG_ERR, "Error accessing firmware file: %s. %m"IPR_EOL, ucode_file);
                    close(dev_fd);
                    close(dasd_ucode_fd);
                    continue;
                }

                /* Map the image in memory */
                p_dasd_image_hdr = mmap(NULL, ucode_stats.st_size,
                                        PROT_READ, MAP_SHARED, dasd_ucode_fd, 0);

                if (p_dasd_image_hdr == MAP_FAILED)
                {
                    syslog(LOG_ERR, "Error reading firmware file: %s. %m"IPR_EOL, ucode_file);
                    close(dev_fd);
                    close(dasd_ucode_fd);
                    continue;
                }

                if ((memcmp(modification_level, dasd_page3_inq.release_level, 4) > 0) ||
                    (force_download == 1) ||
                    (force_devs == 1))
                {
                    if (device_update_blocked)
                    {
                        if (is_af_blocked(&p_cur_ioa->dev[j], 1))
                        {
                            u320_disabled |= (1 << p_device->p_resource_entry->resource_address.bus) |
                                IPR_SAVE_LIMITED_CONFIG;
                        }

                        syslog(LOG_ERR,"Device update to %02X%02X%02X%02X blocked, device in use."IPR_EOL,
                               p_device->p_resource_entry->host_no,
                               p_device->p_resource_entry->resource_address.bus,
                               p_device->p_resource_entry->resource_address.target,
                               p_device->p_resource_entry->resource_address.lun);
                        continue;
                    }

                    if ((p_device->p_resource_entry->subtype == IPR_SUBTYPE_GENERIC_SCSI) &&
                        (memcmp(p_device->p_resource_entry->std_inq_data.vpids.vendor_id, "IBMAS400", 8) == 0))
                    {
                        /* updating AS400 drives which are 512 formatted is not a supported operation */
                        continue;
                    }

                    if (isprint(modification_level[0]) &&
                        isprint(modification_level[1]) &&
                        isprint(modification_level[2]) &&
                        isprint(modification_level[3]) &&
                        isprint(dasd_page3_inq.release_level[0]) &&
                        isprint(dasd_page3_inq.release_level[1]) &&
                        isprint(dasd_page3_inq.release_level[2]) &&
                        isprint(dasd_page3_inq.release_level[3]))
                    {
                        syslog(LOG_NOTICE, "Updating DASD firmware on %02X%02X%02X%02X, using file %s from version %c%c%c%c to %c%c%c%c"IPR_EOL,
                               p_device->p_resource_entry->host_no,
                               p_device->p_resource_entry->resource_address.bus,
                               p_device->p_resource_entry->resource_address.target,
                               p_device->p_resource_entry->resource_address.lun,
                               ucode_file,
                               dasd_page3_inq.release_level[0],
                               dasd_page3_inq.release_level[1],
                               dasd_page3_inq.release_level[2],
                               dasd_page3_inq.release_level[3],
                               modification_level[0], modification_level[1],
                               modification_level[2], modification_level[3]);
                    }
                    else
                    {
                        syslog(LOG_NOTICE, "Updating DASD firmware on %02X%02X%02X%02X, using file %s"IPR_EOL,
                               p_device->p_resource_entry->host_no,
                               p_device->p_resource_entry->resource_address.bus,
                               p_device->p_resource_entry->resource_address.target,
                               p_device->p_resource_entry->resource_address.lun,
                               ucode_file);
                    }

                    p_dasd_image = (void *)p_dasd_image_hdr + image_offset;
                    image_size = ucode_stats.st_size - image_offset;

                    /* Do the update if the firmware file is newer than the firmware loaded on the DASD */
                    if (p_device->p_resource_entry->subtype == IPR_SUBTYPE_GENERIC_SCSI) 
                    {
                        memset(cdb, 0, IPR_CCB_CDB_LEN);

                        /* Issue mode sense to get the block size */
                        cdb[0] = WRITE_BUFFER;
                        cdb[1] = 5; /* Mode 5 */
                        cdb[6] = (image_size & 0xff0000) >> 16;
                        cdb[7] = (image_size & 0x00ff00) >> 8;
                        cdb[8] = image_size & 0x0000ff;

                        rc = sg_ioctl(dev_fd, cdb,
                                      p_dasd_image, image_size, SG_DXFER_TO_DEV,
                                      &sense_data, 120);            
                    }
                    else
                    {
                        memset(&ioa_cmd, 0, sizeof(ioa_cmd));
                        ioa_cmd.buffer = (void *)p_dasd_image;
                        ioa_cmd.buffer_len = image_size;
                        ioa_cmd.read_not_write = 0;
                        ioa_cmd.device_cmd = 1;
                        ioa_cmd.resource_address = p_device->p_resource_entry->resource_address;

                        rc = ipr_ioctl(dev_fd, WRITE_BUFFER, &ioa_cmd);
                    }

                    if (rc != 0)
                        syslog(LOG_ERR, "Write buffer to %02X%02X%02X%02X failed. %m"IPR_EOL,
                               p_device->p_resource_entry->host_no,
                               p_device->p_resource_entry->resource_address.bus,
                               p_device->p_resource_entry->resource_address.target,
                               p_device->p_resource_entry->resource_address.lun);
                }

                close(dev_fd);
                close(dasd_ucode_fd);

                if (munmap(p_dasd_image_hdr, ucode_stats.st_size))
                    syslog(LOG_ERR, "munmap failed. %m"IPR_EOL);
            }
            else
            {
                syslog(LOG_ERR, "Cannot open device %s. Retry failed. %m"IPR_EOL, p_dev_name);
                continue;
            }
        }

        munmap(p_image_hdr, ucode_stats.st_size);
        close (ucode_fd);
        close (fd);

        if (RC_SUCCESS == ipr_config_file_valid(config_file_name))
        {
            ipr_set_page_28(p_cur_ioa, u320_disabled,
                            do_download | force_download | force_ioas);
        }
        else
        {
            ipr_set_page_28_init(p_cur_ioa, u320_disabled);
        }
    }

    return rc;
}

int open_dev(struct ipr_ioa *p_ioa, struct ipr_device *p_device)
{
    int dev_fd;

    if (p_device->p_resource_entry->subtype == IPR_SUBTYPE_GENERIC_SCSI)
    {
        dev_fd = open(p_device->gen_name, O_RDWR);

        if ((dev_fd <= 1) && ((errno == ENOENT) || (errno == ENXIO)))
        {
            syslog(LOG_ERR, "Cannot open device %s. %m"IPR_EOL, p_device->gen_name);
            syslog(LOG_ERR, "Rescanning SCSI bus for device"IPR_EOL);
            scan_device(p_device->p_resource_entry->resource_address, p_ioa->host_num);

            dev_fd = open(p_device->gen_name, O_RDWR);
        }
    }
    else
        dev_fd = open(p_ioa->ioa.dev_name, O_RDWR);

    return dev_fd;
}

